@@ -38,7 +38,7 @@ typedef struct _gss_krb5_ctx_struct
   Shishi *sh;
   Shishi_ap *ap;
   Shishi_tkt *tkt;
-  Shishi_tkt *key;
+  Shishi_key *key;
   int acceptor;
   int acceptseqnr;
   int initseqnr;
@@ -52,6 +52,11 @@ typedef struct _gss_krb5_ctx_struct
 #define TOK_MIC    "\x01\x01"
 #define TOK_WRAP   "\x02\x01"
 
+#define C2I(buf) ((buf[0] & 0xFF) |		\
+		  ((buf[1] & 0xFF) << 8) |	\
+		  ((buf[2] & 0xFF) << 16) |	\
+		  ((buf[3] & 0xFF) << 24))
+
 static void
 hexprint (const unsigned char *str, int len)
 {
@@ -100,6 +105,9 @@ gss_krb5_init_sec_context (OM_uint32 * minor_status,
   if (!output_token)
     return GSS_S_FAILURE;
 
+  if (actual_mech_type)
+    (*actual_mech_type) = GSS_KRB5;
+
   /* XXX mech_type not tested */
 
   if (*context_handle == GSS_C_NO_CONTEXT)
@@ -113,9 +121,14 @@ gss_krb5_init_sec_context (OM_uint32 * minor_status,
 
       *context_handle = ctx;
 
-      rc = shishi_init(&h);
-      if (rc != SHISHI_OK)
-	return GSS_S_FAILURE;
+      if (initiator_cred_handle)
+	h = initiator_cred_handle->krb5->sh;
+      else
+	{
+	  rc = shishi_init(&h);
+	  if (rc != SHISHI_OK)
+	    return GSS_S_FAILURE;
+	}
       ctx->krb5->sh = h;
 
       ctx->peerptr = &ctx->peer;
@@ -142,9 +155,12 @@ gss_krb5_init_sec_context (OM_uint32 * minor_status,
 	  Shishi_tkts_hint hint;
 
 	  memset(&hint, 0, sizeof(hint));
-	  hint.server = ctx->peer.value;
+	  hint.server = malloc(ctx->peer.length + 1);
+	  memcpy(hint.server, ctx->peer.value, ctx->peer.length);
+	  hint.server[ctx->peer.length] = '\0';
 
 	  tkt = shishi_tkts_get (shishi_tkts_default (h), &hint);
+	  free(hint.server);
 	  if (!tkt)
 	    return GSS_S_FAILURE;
 
@@ -346,8 +362,8 @@ gss_krb5_wrap (OM_uint32 * minor_status,
 	char header[8];
 	char encseqno[8];
 	char seqno[8];
-	char eseqno[8];
-	char cksum[8];
+	char *eseqno;
+	char *cksum;
 	char confounder[8];
 	char tmp[20];
 	char *pt;
@@ -384,14 +400,13 @@ gss_krb5_wrap (OM_uint32 * minor_status,
 	memset (data.value + 16 + input_message_buffer->length,
 		padlength, padlength);
 
-	tmplen = 8;
 	rc = shishi_checksum (k5->sh,
 			      k5->key,
-			      0, 42,
+			      0, SHISHI_RSA_MD5_DES_GSS,
 			      data.value,
 			      16 + input_message_buffer->length + padlength,
-			      cksum, &tmplen);
-	if (rc != SHISHI_OK)
+			      &cksum, &tmplen);
+	if (rc != SHISHI_OK || tmplen != 8)
 	  return GSS_S_FAILURE;
 
 	/* seq_nr */
@@ -401,20 +416,21 @@ gss_krb5_wrap (OM_uint32 * minor_status,
 	seqno[3] = k5->initseqnr >> 24 & 0xFF;
 	memset(seqno + 4, k5->acceptor ? 0xFF : 0, 4);
 
-	tmplen = 8;
 	rc = shishi_encrypt_iv_etype(k5->sh,
 				     k5->key,
 				     0, SHISHI_DES_CBC_NONE,
 				     cksum, 8, /* cksum */
 				     seqno, 8,
-				     eseqno, &tmplen);
+				     &eseqno, &tmplen);
 	if (rc != SHISHI_OK || tmplen != 8)
 	  return GSS_S_FAILURE;
 
 	/* put things in place */
 	memcpy (data.value, header, 8);
 	memcpy (data.value + 8, eseqno, 8);
+	free (eseqno);
 	memcpy (data.value + 16, cksum, 8);
+	free (cksum);
 	memcpy (data.value + 24, confounder, 8);
 	memcpy (data.value + 32, input_message_buffer->value,
 		input_message_buffer->length);
@@ -429,65 +445,70 @@ gss_krb5_wrap (OM_uint32 * minor_status,
       break;
 
     case SHISHI_DES3_CBC_HMAC_SHA1_KD:
+      {
+	char *tmp;
 
-      padlength = 8 - input_message_buffer->length % 8;
-      data.length = 8 + 8 + 20 + 8 + input_message_buffer->length + padlength;
-      data.value = xmalloc(data.length);
+	padlength = 8 - input_message_buffer->length % 8;
+	data.length = 8 + 8 + 20 + 8 + input_message_buffer->length + padlength;
+	data.value = xmalloc(data.length);
 
-      /* XXX encrypt data iff confidential option chosen */
+	/* XXX encrypt data iff confidential option chosen */
 
-      /* Compute checksum over header, confounder, input string, and pad */
+	/* Compute checksum over header, confounder, input string, and pad */
 
-      memcpy (data.value, TOK_WRAP, 2);	       /* TOK_ID: Wrap */
-      memcpy (data.value + 2, "\x04\x00", 2);  /* SGN_ALG: 3DES */
-      memcpy (data.value + 4, "\xFF\xFF", 2);  /* SEAL_ALG: none */
-      memcpy (data.value + 6, "\xFF\xFF", 2);  /* filler */
-      rc = shishi_randomize(k5->sh, data.value + 8, 8);
-      if (rc != SHISHI_OK)
-	return GSS_S_FAILURE;
-      memcpy (data.value + 16, input_message_buffer->value,
-	      input_message_buffer->length);
-      memset (data.value + 16 + input_message_buffer->length,
-	      padlength, padlength);
+	memcpy (data.value, TOK_WRAP, 2);	       /* TOK_ID: Wrap */
+	memcpy (data.value + 2, "\x04\x00", 2);  /* SGN_ALG: 3DES */
+	memcpy (data.value + 4, "\xFF\xFF", 2);  /* SEAL_ALG: none */
+	memcpy (data.value + 6, "\xFF\xFF", 2);  /* filler */
+	rc = shishi_randomize(k5->sh, data.value + 8, 8);
+	if (rc != SHISHI_OK)
+	  return GSS_S_FAILURE;
+	memcpy (data.value + 16, input_message_buffer->value,
+		input_message_buffer->length);
+	memset (data.value + 16 + input_message_buffer->length,
+		padlength, padlength);
 
-      tmplen = 20;
-      rc = shishi_checksum (k5->sh,
-			    k5->key,
-			    SHISHI_KEYUSAGE_GSS_R2, SHISHI_HMAC_SHA1_DES3_KD,
-			    data.value,
-			    16 + input_message_buffer->length + padlength,
-			    data.value + 16, &tmplen);
-      if (rc != SHISHI_OK)
-	return GSS_S_FAILURE;
+	rc = shishi_checksum (k5->sh,
+			      k5->key,
+			      SHISHI_KEYUSAGE_GSS_R2, SHISHI_HMAC_SHA1_DES3_KD,
+			      data.value,
+			      16 + input_message_buffer->length + padlength,
+			      &tmp, &tmplen);
+	if (rc != SHISHI_OK || tmplen != 20)
+	  return GSS_S_FAILURE;
 
-      memcpy (data.value + 36, data.value + 8, 8);
+	memcpy (data.value + 16, tmp, tmplen);
+	memcpy (data.value + 36, data.value + 8, 8);
 
-      /* seq_nr */
-      ((char*)data.value + 8)[0] = k5->initseqnr & 0xFF;
-      ((char*)data.value + 8)[1] = k5->initseqnr >> 8 & 0xFF;
-      ((char*)data.value + 8)[2] = k5->initseqnr >> 16 & 0xFF;
-      ((char*)data.value + 8)[3] = k5->initseqnr >> 24 & 0xFF;
-      memset(data.value + 8 + 4, k5->acceptor ? 0xFF : 0, 4);
-      tmplen = 8;
-      rc = shishi_encrypt_iv_etype(k5->sh,
-				   k5->key,
-				   0, SHISHI_DES3_CBC_NONE,
-				   data.value + 16, 8, /* cksum */
-				   data.value + 8, 8,
-				   data.value + 8, &tmplen);
-      if (rc != SHISHI_OK || tmplen != 8)
-	return GSS_S_FAILURE;
+	/* seq_nr */
+	((char*)data.value + 8)[0] = k5->initseqnr & 0xFF;
+	((char*)data.value + 8)[1] = k5->initseqnr >> 8 & 0xFF;
+	((char*)data.value + 8)[2] = k5->initseqnr >> 16 & 0xFF;
+	((char*)data.value + 8)[3] = k5->initseqnr >> 24 & 0xFF;
+	memset(data.value + 8 + 4, k5->acceptor ? 0xFF : 0, 4);
 
-      memcpy (data.value + 8 + 8 + 20 + 8, input_message_buffer->value,
-	      input_message_buffer->length);
-      memset (data.value + 8 + 8 + 20 + 8 + input_message_buffer->length,
-	      padlength, padlength);
+	rc = shishi_encrypt_iv_etype(k5->sh,
+				     k5->key,
+				     0, SHISHI_DES3_CBC_NONE,
+				     data.value + 16, 8, /* cksum */
+				     data.value + 8, 8,
+				     &tmp, &tmplen);
+	if (rc != SHISHI_OK || tmplen != 8)
+	  return GSS_S_FAILURE;
 
-      rc = gss_encapsulate_token(&data, GSS_KRB5, output_message_buffer);
-      if (!rc)
-	return GSS_S_FAILURE;
-      k5->initseqnr++;
-      break;
+	memcpy(data.value + 8, tmp, tmplen);
+	free (tmp);
+	memcpy (data.value + 8 + 8 + 20 + 8, input_message_buffer->value,
+		input_message_buffer->length);
+	memset (data.value + 8 + 8 + 20 + 8 + input_message_buffer->length,
+		padlength, padlength);
+
+	rc = gss_encapsulate_token(&data, GSS_KRB5, output_message_buffer);
+	if (!rc)
+	  return GSS_S_FAILURE;
+	k5->initseqnr++;
+	break;
+      }
 
     default:
       return GSS_S_FAILURE;
@@ -548,7 +569,7 @@ gss_krb5_unwrap (OM_uint32 * minor_status,
 	char seqno[8];
 	char cksum[8];
 	char confounder[8];
-	char tmp[20];
+	char *tmp;
 	size_t cksumlen = 8;
 	int seqnr;
 	int i;
@@ -574,31 +595,24 @@ gss_krb5_unwrap (OM_uint32 * minor_status,
 
 	/* XXX decrypt data iff confidential option chosen */
 
-	/* XXX here we rely on Shishi to decrypt the data into the
-	   output buffer before verifying it.  The verification will
-	   always fail (RFC 1964 do not use real des-cbc-md5) so we
-	   ignore the return code too.  We check the decrypted data
-	   against the sequence number, so this isn't evil.  Correct
-	   solution: add a GSS-DES cipher suite to Shishi.  Or use
-	   libgcrypt here in GSS, there really isn't anything Kerberos
-	   5 related about the encryption operation used by RFC 1964
-	   so it doesn't fit into Shishi. */
-	i = 8;
 	rc = shishi_decrypt_iv_etype (k5->sh,
 				      k5->key,
-				      0, SHISHI_DES_CBC_MD5,
+				      0, SHISHI_DES_CBC_NONE,
 				      cksum, 8,
 				      encseqno, 8,
-				      tmp, &i);
-	memcpy(seqno, tmp, 8);
+				      &tmp, &i);
+	if (rc != SHISHI_OK)
+	  return GSS_S_FAILURE;
 	if (i != 8)
 	  return GSS_S_BAD_MIC;
+	memcpy(seqno, tmp, 8);
+	free (tmp);
 
 	if (memcmp(seqno + 4, k5->acceptor ? "\x00\x00\x00\x00" :
 		   "\xFF\xFF\xFF\xFF", 4) != 0)
 	  return GSS_S_BAD_MIC;
 
-	seqnr = (seqno[0]|(seqno[1] << 8)|(seqno[2] << 16)|(seqno[3] << 24));
+	seqnr = C2I(seqno);
 	if (seqnr != k5->acceptseqnr)
 	  return GSS_S_BAD_MIC;
 
@@ -617,15 +631,16 @@ gss_krb5_unwrap (OM_uint32 * minor_status,
 	memcpy(data.value + 24, confounder, 8);
 
 	/* Checksum header + confounder + data + pad */
-	tmplen = 8;
 	rc = shishi_checksum (k5->sh,
 			      k5->key,
-			      0, 42,
+			      0, SHISHI_RSA_MD5_DES_GSS,
 			      data.value + 16, data.length - 16,
-			      data.value + 8, &tmplen);
-	if (rc != SHISHI_OK)
+			      &tmp, &tmplen);
+	if (rc != SHISHI_OK || tmplen != 8)
 	  return GSS_S_FAILURE;
 
+	memcpy (data.value + 8, tmp, tmplen);
+
 	/* Compare checksum */
 	if (tmplen != 8 || memcmp (cksum, data.value + 8, 8) != 0)
 	  return GSS_S_BAD_MIC;
@@ -641,6 +656,7 @@ gss_krb5_unwrap (OM_uint32 * minor_status,
       {
 	size_t padlen;
 	unsigned char *p;
+	char *t;
 	char cksum[20];
 	size_t cksumlen = 20;
 	int i;
@@ -652,21 +668,23 @@ gss_krb5_unwrap (OM_uint32 * minor_status,
 
 	/* XXX decrypt data iff confidential option chosen */
 
-	i=20;
 	p = data.value + 8;
 	rc = shishi_decrypt_iv_etype (k5->sh,
 				      k5->key,
 				      0, SHISHI_DES3_CBC_NONE,
 				      cksum, 8,
 				      p, 8,
-				      p, &i);
-	if (rc != SHISHI_OK)
+				      &t, &i);
+	if (rc != SHISHI_OK || i != 20)
 	  return GSS_S_FAILURE;
 
+	memcpy(p, t, i);
+	free (t);
+
 	if (memcmp(p + 4, k5->acceptor ? "\x00\x00\x00\x00" :
 		   "\xFF\xFF\xFF\xFF", 4) != 0)
 	  return GSS_S_BAD_MIC;
-	if ((p[0]|(p[1] << 8)|(p[2] << 16)|(p[3] << 24)) != k5->acceptseqnr)
+	if (C2I(p) != k5->acceptseqnr)
 	  return GSS_S_BAD_MIC;
 
 	k5->acceptseqnr++;
@@ -683,15 +701,17 @@ gss_krb5_unwrap (OM_uint32 * minor_status,
 	memcpy(data.value + 8 + 20, data.value, 8);
 
 	/* Checksum header + confounder + data + pad */
-	tmplen = 20;
 	rc = shishi_checksum (k5->sh,
 			      k5->key,
 			      SHISHI_KEYUSAGE_GSS_R2, SHISHI_HMAC_SHA1_DES3_KD,
 			      data.value + 20 + 8, data.length - 20 - 8,
-			      data.value + 8 + 8, &tmplen);
-	if (rc != SHISHI_OK)
+			      &t, &tmplen);
+	if (rc != SHISHI_OK || tmplen != 20)
 	  return GSS_S_FAILURE;
 
+	memcpy(data.value + 8 + 8, t, tmplen);
+	free (t);
+
 	/* Compare checksum */
 	if (tmplen != 20 || memcmp (cksum, data.value + 8 + 8, 20) != 0)
 	  return GSS_S_BAD_MIC;
@@ -829,7 +849,7 @@ gss_krb5_acquire_cred1 (OM_uint32 * minor_status,
       buf.length = strlen(buf.value);
       maj_stat = gss_import_name (minor_status, &buf,
 				  GSS_C_NT_HOSTBASED_SERVICE,
-				  &desired_name);
+				  (gss_name_t*)&desired_name);
       if (GSS_ERROR(maj_stat))
 	return maj_stat;
     }
@@ -902,6 +922,7 @@ gss_krb5_acquire_cred (OM_uint32 * minor_status,
     }
 
   p = xcalloc(sizeof(*p), 1);
+  p->mech = GSS_KRB5;
   p->krb5 = xcalloc(sizeof(*p->krb5), 1);
   p->krb5->peerptr = &p->krb5->peer;
 
@@ -927,6 +948,64 @@ gss_krb5_acquire_cred (OM_uint32 * minor_status,
   return GSS_S_COMPLETE;
 }
 
+OM_uint32
+gss_krb5_inquire_cred (OM_uint32 * minor_status,
+		       const gss_cred_id_t cred_handle,
+		       gss_name_t * name,
+		       OM_uint32 * lifetime,
+		       gss_cred_usage_t * cred_usage,
+		       gss_OID_set * mechanisms)
+{
+  OM_uint32 maj_stat;
+
+  if (minor_status)
+    *minor_status = 0;
+
+  if (cred_handle == GSS_C_NO_CREDENTIAL)
+    {
+      
+    }
+
+  if (name)
+    {
+      maj_stat = gss_duplicate_name (minor_status, cred_handle->krb5->peerptr,
+				     name);
+      if (GSS_ERROR (maj_stat))
+	return maj_stat;
+    }
+
+  if (lifetime)
+    {
+      if (cred_handle->krb5->tkt)
+	{
+	  time_t end = shishi_tkt_endctime (cred_handle->krb5->tkt);
+	  time_t now = time(NULL);
+
+	  if (shishi_tkt_valid_now_p (cred_handle->krb5->tkt))
+	    *lifetime = (OM_uint32) difftime (now, end);
+	  else
+	    *lifetime = 0;
+	}
+      else
+	*lifetime = GSS_C_INDEFINITE;
+    }
+
+  if (cred_usage)
+    *cred_usage = GSS_C_BOTH;
+
+  if (mechanisms)
+    {
+      maj_stat = gss_create_empty_oid_set (minor_status, mechanisms);
+      if (GSS_ERROR (maj_stat))
+	return maj_stat;
+      maj_stat = gss_add_oid_set_member (minor_status, GSS_KRB5, mechanisms);
+      if (GSS_ERROR (maj_stat))
+	return maj_stat;
+    }
+
+  return GSS_S_COMPLETE;
+}
+
 OM_uint32
 gss_krb5_accept_sec_context (OM_uint32 * minor_status,
 			     gss_ctx_id_t * context_handle,
