@@ -28,11 +28,16 @@ main (int argc, char *argv[])
   char buffer2[BUFSIZ];
   char *p, *q;
   int n, res;
-  gss_uint32 maj_stat, min_stat;
+  gss_uint32 maj_stat, min_stat, msgctx;
   gss_buffer_desc bufdesc, bufdesc2;
   gss_name_t service;
   gss_OID_set oids;
 
+#if HAVE_LOCALE_H
+  setlocale (LC_ALL, "");
+#endif
+  bindtextdomain (PACKAGE, LOCALEDIR);
+
   do
     if (strcmp (argv[argc - 1], "-v") == 0 ||
 	strcmp (argv[argc - 1], "--verbose") == 0)
@@ -312,6 +317,19 @@ main (int argc, char *argv[])
     fail("gss_release_oid_set() failed (%d,%d)\n", maj_stat, min_stat);
 #endif
 
+  /* Check display_status */
+  msgctx = 0;
+  maj_stat = gss_display_status (&min_stat, GSS_S_COMPLETE, GSS_C_GSS_CODE,
+				 GSS_C_NO_OID, &msgctx, &bufdesc);
+  if (maj_stat == GSS_S_COMPLETE)
+    success("gss_display_status() OK\n");
+  else
+    fail("gss_display_status() failed (%d,%d)\n", maj_stat, min_stat);
+
+  if (debug)
+    printf("    Display status for GSS_S_COMPLETE => %*s\n",
+	   bufdesc.length, bufdesc.value);
+
   if (verbose)
     printf ("Name self tests done with %d errors\n", error_count);
 
